"""
Reimplementation of several features in the emcee and lmfit packages, in order to make it work correctly.

.. moduleauthor:: Wouter Gins <wouter.a.gins@jyu.fi>
"""
import h5py

from collections import namedtuple
from copy import deepcopy
import inspect
import multiprocessing
import numbers
import warnings

import numpy as np
import uncertainties

import emcee
from emcee.autocorr import AutocorrError, integrated_time
from lmfit import Minimizer
import lmfit.minimizer
from typing import Dict, List, Optional, Union

try:
    import pandas as pd
    from pandas import isnull
    HAS_PANDAS = True
except ImportError:
    HAS_PANDAS = False
    isnull = np.isnan
_make_random_gen = lmfit.minimizer._make_random_gen
isnull = lmfit.minimizer.isnull
_nan_policy = lmfit.minimizer._nan_policy

__all__ = ['SATLASSampler', 'SATLASHDFBackend', 'SATLASMinimizer', 'minimize']

def ndarray_to_list_of_dicts(
    x: np.ndarray, key_map: Dict[str, Union[int, List[int]]]
) -> List[Dict[str, Union[np.number, np.ndarray]]]:
    """
    A helper function to convert a ``np.ndarray`` into a list
    of dictionaries of parameters. Used when parameters are named.
    Args:
      x (np.ndarray): parameter array of shape ``(N, n_dim)``, where
        ``N`` is an integer
      key_map (Dict[str, Union[int, List[int]]):
    Returns:
      list of dictionaries of parameters
    """
    return [{key: xi[val] for key, val in key_map.items()} for xi in x]

class SATLASSampler(emcee.EnsembleSampler):
    def compute_log_prob(self, coords):
        """Calculate the vector of log-probability for the walkers

        Args:
            coords: (ndarray[..., ndim]) The position vector in parameter
                space where the probability should be calculated.

        This method returns:

        * log_prob: A vector of log-probabilities with one entry for each
          walker in this sub-ensemble.
        * blob: The list of meta data returned by the ``log_post_fn`` at
          this position or ``None`` if nothing was returned.

        """
        p = coords

        # Check that the parameters are in physical ranges.
        if np.any(np.isinf(p)):
            raise ValueError("At least one parameter value was infinite")
        if np.any(np.isnan(p)):
            raise ValueError("At least one parameter value was NaN")

        # If the parmaeters are named, then switch to dictionaries
        if self.params_are_named:
            p = ndarray_to_list_of_dicts(p, self.parameter_names)

        # Run the log-probability calculations (optionally in parallel).
        if self.vectorize:
            results = self.log_prob_fn(p)
        else:
            # If the `pool` property of the sampler has been set (i.e. we want
            # to use `multiprocessing`), use the `pool`'s map method.
            # Otherwise, just use the built-in `map` function.
            if self.pool is not None:
                map_func = self.pool.map
            else:
                map_func = map
            results = list(map_func(self.log_prob_fn, p))

        log_prob = np.array([float(l) for l in results])
        blob = None

        # Check for log_prob returning NaN.
        if np.any(np.isnan(log_prob)):
            raise ValueError("Probability function returned NaN")

        return log_prob, blob

class SATLASHDFBackend(emcee.backends.HDFBackend):
    @property
    def labels(self):
        with self.open() as f:
            g = f[self.name]
            return g.attrs["labels"]
    
    @labels.setter
    def labels(self, labels):
        with self.open("a") as f:
            g = f[self.name]
            g.attrs["labels"] = labels


class SATLASMinimizer(Minimizer):

    def process_walk(self, params, chain):
        result = self.prepare_fit(params)
        params = result.params
        nvarys = result.nvarys
        result.method = 'emcee'

        flatchain = chain.reshape((-1, nvarys))
        steps = chain.shape[0]
        nwalkers = chain.shape[1]
        quantiles = np.percentile(flatchain, [15.87, 50, 84.13], axis=0)

        for i, var_name in enumerate(result.var_names):
            std_l, median, std_u = quantiles[:, i]
            params[var_name].value = median
            params[var_name].stderr = 0.5 * (std_u - std_l)
            params[var_name].correl = {}

        params.update_constraints()

        # work out correlation coefficients
        corrcoefs = np.corrcoef(flatchain.T)

        for i, var_name in enumerate(result.var_names):
            for j, var_name2 in enumerate(result.var_names):
                if i != j:
                    result.params[var_name].correl[var_name2] = corrcoefs[i, j]

        result.errorbars = True
        result.nvarys = len(result.var_names)
        result.nfev = nwalkers*steps

        try:
            result.acor = emcee.autocorr.integrated_time(chain)
        except AutocorrError as e:
            print(str(e))
        return result

        # Calculate the residual with the "best fit" parameters
        

    def emcee(self, params=None, steps=1000, nwalkers=100, burn=0, thin=1,
              ntemps=1, pos=None, reuse_sampler=False, workers=1,
              float_behavior='posterior', is_weighted=True, seed=None,
              progress=True, mcmc_kwargs={}, sampler_kwargs={}, sampler=emcee.EnsembleSampler):
        if ntemps > 1:
            msg = ("'ntemps' has no effect anymore, since the PTSampler was "
                   "removed from emcee version 3.")
            raise DeprecationWarning(msg)

        tparams = params
        # if you're reusing the sampler then nwalkers have to be
        # determined from the previous sampling
        if reuse_sampler:
            if not hasattr(self, 'sampler') or not hasattr(self, '_lastpos'):
                raise ValueError("You wanted to use an existing sampler, but "
                                 "it hasn't been created yet")
            if len(self._lastpos.shape) == 2:
                nwalkers = self._lastpos.shape[0]
            elif len(self._lastpos.shape) == 3:
                nwalkers = self._lastpos.shape[1]
            tparams = None

        result = self.prepare_fit(params=tparams)
        params = result.params

        # check if the userfcn returns a vector of residuals
        out = self.userfcn(params, *self.userargs, **self.userkws)
        out = np.asarray(out).ravel()
        if out.size > 1 and is_weighted is False:
            # we need to marginalise over a constant data uncertainty
            if '__lnsigma' not in params:
                # __lnsigma should already be in params if is_weighted was
                # previously set to True.
                params.add('__lnsigma', value=0.01, min=-np.inf, max=np.inf,
                           vary=True)
                # have to re-prepare the fit
                result = self.prepare_fit(params)
                params = result.params

        result.method = 'emcee'

        # Removing internal parameter scaling. We could possibly keep it,
        # but I don't know how this affects the emcee sampling.
        bounds = []
        var_arr = np.zeros(len(result.var_names))
        i = 0
        for par in params:
            param = params[par]
            if param.expr is not None:
                param.vary = False
            if param.vary:
                var_arr[i] = param.value
                i += 1
            else:
                # don't want to append bounds if they're not being varied.
                continue

            param.from_internal = lambda val: val
            lb, ub = param.min, param.max
            if lb is None or lb is np.nan:
                lb = -np.inf
            if ub is None or ub is np.nan:
                ub = np.inf
            bounds.append((lb, ub))
        bounds = np.array(bounds)

        self.nvarys = len(result.var_names)

        # set up multiprocessing options for the samplers
        auto_pool = None
        # sampler_kwargs = {}
        if isinstance(workers, int) and workers > 1 and HAS_DILL:
            auto_pool = multiprocessing.Pool(workers)
            sampler_kwargs['pool'] = auto_pool
        elif hasattr(workers, 'map'):
            sampler_kwargs['pool'] = workers

        # function arguments for the log-probability functions
        # these values are sent to the log-probability functions by the sampler.
        lnprob_args = (self.userfcn, params, result.var_names, bounds)
        lnprob_kwargs = {'is_weighted': is_weighted,
                         'float_behavior': float_behavior,
                         'userargs': self.userargs,
                         'userkws': self.userkws,
                         'nan_policy': self.nan_policy}

        sampler_kwargs['args'] = lnprob_args
        sampler_kwargs['kwargs'] = lnprob_kwargs

        # set up the random number generator
        rng = _make_random_gen(seed)

        # now initialise the samplers
        if reuse_sampler:
            if auto_pool is not None:
                self.sampler.pool = auto_pool

            p0 = self._lastpos
            if p0.shape[-1] != self.nvarys:
                raise ValueError("You cannot reuse the sampler if the number "
                                 "of varying parameters has changed")

        else:
            p0 = 1 + rng.randn(nwalkers, self.nvarys) * 1.e-4
            p0 *= var_arr
            sampler_kwargs['pool'] = auto_pool
            self.sampler = sampler(nwalkers, self.nvarys,
                                                 self._lnprob, **sampler_kwargs)
        backend = sampler_kwargs.pop('backend')
        if backend is not None:
            backend.reset(nwalkers, self.nvarys)
            backend.labels = result.var_names

        # user supplies an initialisation position for the chain
        # If you try to run the sampler with p0 of a wrong size then you'll get
        # a ValueError. Note, you can't initialise with a position if you are
        # reusing the sampler.
        if pos is not None and not reuse_sampler:
            tpos = np.asfarray(pos)
            if p0.shape == tpos.shape:
                pass
            # trying to initialise with a previous chain
            elif tpos.shape[-1] == self.nvarys:
                tpos = tpos[-1]
            else:
                raise ValueError('pos should have shape (nwalkers, nvarys)')
            p0 = tpos

        # if you specified a seed then you also need to seed the sampler
        if seed is not None:
            self.sampler.random_state = rng.get_state()

        # now do a production run, sampling all the time
        try:
            output = self.sampler.run_mcmc(p0, steps, progress=progress, **mcmc_kwargs)
            self._lastpos = output.coords
        except AbortFitException:
            result.aborted = True
            result.message = "Fit aborted by user callback. Could not estimate error-bars."
            result.success = False
            result.nfev = self.result.nfev
            output = None

        # discard the burn samples and thin
        chain = self.sampler.get_chain(thin=thin, discard=burn)[..., :, :]
        lnprobability = self.sampler.get_log_prob(thin=thin, discard=burn)[..., :]
        flatchain = chain.reshape((-1, self.nvarys))
        if not result.aborted:
            quantiles = np.percentile(flatchain, [15.87, 50, 84.13], axis=0)

            for i, var_name in enumerate(result.var_names):
                std_l, median, std_u = quantiles[:, i]
                params[var_name].value = median
                params[var_name].stderr = 0.5 * (std_u - std_l)
                params[var_name].correl = {}

            params.update_constraints()

            # work out correlation coefficients
            corrcoefs = np.corrcoef(flatchain.T)

            for i, var_name in enumerate(result.var_names):
                for j, var_name2 in enumerate(result.var_names):
                    if i != j:
                        result.params[var_name].correl[var_name2] = corrcoefs[i, j]

        result.chain = np.copy(chain)
        result.lnprob = np.copy(lnprobability)
        result.errorbars = True
        result.nvarys = len(result.var_names)
        result.nfev = nwalkers*steps

        try:
            result.acor = self.sampler.get_autocorr_time()
        except AutocorrError as e:
            print(str(e))
        result.acceptance_fraction = self.sampler.acceptance_fraction

        # Calculate the residual with the "best fit" parameters
        out = self.userfcn(params, *self.userargs, **self.userkws)
        result.residual = _nan_policy(out, nan_policy=self.nan_policy,
                                      handle_inf=False)

        # If uncertainty was automatically estimated, weight the residual properly
        if (not is_weighted) and (result.residual.size > 1):
            if '__lnsigma' in params:
                result.residual = result.residual/np.exp(params['__lnsigma'].value)

        # Calculate statistics for the two standard cases:
        if isinstance(result.residual, np.ndarray) or (float_behavior == 'chi2'):
            result._calculate_statistics()

        # Handle special case unique to emcee:
        # This should eventually be moved into result._calculate_statistics.
        elif float_behavior == 'posterior':
            result.ndata = 1
            result.nfree = 1

            # assuming prior prob = 1, this is true
            _neg2_log_likel = -2*result.residual

            # assumes that residual is properly weighted, avoid overflowing np.exp()
            result.chisqr = np.exp(min(650, _neg2_log_likel))

            result.redchi = result.chisqr / result.nfree
            result.aic = _neg2_log_likel + 2 * result.nvarys
            result.bic = _neg2_log_likel + np.log(result.ndata) * result.nvarys

        if auto_pool is not None:
            auto_pool.terminate()

        return result


def minimize(fcn, params, method='leastsq', args=None, kws=None, iter_cb=None,
             scale_covar=True, nan_policy='raise', reduce_fcn=None,
             calc_covar=True, max_nfev=None, **fit_kws):
    minimizer_kws = fit_kws.pop('minimizer_kws', {})
    fitter = SATLASMinimizer(fcn, params, fcn_args=args, fcn_kws=kws,
                       iter_cb=iter_cb, scale_covar=scale_covar,
                       nan_policy=nan_policy, reduce_fcn=reduce_fcn,
                       calc_covar=calc_covar, max_nfev=max_nfev, **fit_kws)
    return fitter.minimize(method=method, **minimizer_kws)



"""
Implementation of the base Fitter, Source, Model and Parameter classes

.. moduleauthor:: Wouter Gins <wouter.a.gins@jyu.fi>
"""

import numpy as np
import scipy.optimize as optimize
import lmfit as lm
import emcee
import copy
from .overwrite import SATLASSampler, SATLASHDFBackend, minimize, SATLASMinimizer

__all__ = ['Fitter', 'Source', 'Model', 'Parameter']

class Fitter:
    def __init__(self):
        super().__init__()
        self.sources = []
        self.pars = {}
        self.bounds = optimize.Bounds([], [])
        self.share = []
        self.shareModel = []
        self.priors = []

    def shareParams(self, parameter_name):
        try:
            self.share.extend(parameter_name)
        except:
            self.share.append(parameter_name)

    def shareModelParams(self, parameter_name):
        try:
            self.shareModel.extend(parameter_name)
        except:
            self.shareModel.append(parameter_name)

    def setParamPrior(self, source, model, parameter_name, value, uncertainty):
        self.priors.append((source, model, parameter_name, value, uncertainty))

    def addSource(self, source, name=None):
        if name is None:
            name = source.name
        self.sources.append((name, source))

    def createParameters(self):
        for name, source in self.sources:
            self.pars[name] = source.params()

    def createBounds(self):
        lower = []
        upper = []
        for source_name in self.pars.keys():
            p = self.pars[source_name]
            for model_name in p.keys():
                pars = p[model_name]
                for parameter_name in pars.keys():
                    parameter = pars[parameter_name]
                    if not parameter.vary:
                        l = parameter.value
                        u = parameter.value
                    else:
                        l = parameter.min
                        u = parameter.max
                    lower.append(l)
                    upper.append(u)
        self.bounds = optimize.Bounds(lower, upper)

    def createLmParameters(self):
        lmpars = lm.Parameters()
        sharing = {}
        sharingModel = {}
        tuples = ()
        for source_name in self.pars.keys():
            p = self.pars[source_name]
            for model_name in p.keys():
                pars = p[model_name]
                for parameter_name in pars.keys():
                    parameter = pars[parameter_name]
                    n = '___'.join([source_name, model_name, parameter_name])
                    if parameter_name in self.share:
                        if parameter_name in sharing.keys():
                            expr = sharing[parameter_name]
                        else:
                            sharing[parameter_name] = n
                            expr = parameter.expr
                    elif parameter_name in self.shareModel:
                        if parameter_name in sharingModel.keys() and model_name in sharingModel[parameter_name].keys():
                            expr = sharingModel[parameter_name][model_name]
                        else:
                            try:
                                sharingModel[parameter_name][model_name] = n
                            except:
                                sharingModel[parameter_name] = {model_name: n}
                            expr = parameter.expr
                    else:
                        expr = parameter.expr
                    tuples += ((n, parameter.value, parameter.vary, parameter.min, parameter.max, expr, None),)
        lmpars.add_many(*tuples)
        self.lmpars = lmpars

    def createParameterList(self):
        x = []
        for source_name in self.pars.keys():
            p = self.pars[source_name]
            for model_name in p.keys():
                pars = p[model_name]
                for parameter_name in pars.keys():
                    x.append(pars[parameter_name].value)
        return x

    def f(self):
        f = []
        for name, source in self.sources:
            f.append(source.f())
        return np.hstack(f)

    def y(self):
        y = []
        for _, source in self.sources:
            y.append(source.y)
        return np.hstack(y)

    def yerr(self):
        yerr = []
        for _, source in self.sources:
            yerr.append(source.yerr())
        return np.hstack(yerr)

    def setParameters(self, params):
        for p in params.keys():
            source_name, model_name, parameter_name = p.split('___')
            self.pars[source_name][model_name][parameter_name].value = params[p].value

    def setUncertainties(self, params):
        for p in params.keys():
            source_name, model_name, parameter_name = p.split('___')
            self.pars[source_name][model_name][parameter_name].unc = params[p].stderr

    def resid(self):
        model_calcs = self.f()
        resid = (model_calcs-self.temp_y)/self.yerr()
        if np.any(np.isnan(resid)):
            return np.inf
        else:
            return resid

    def gaussianPriors(self):
        return [(self.pars[source][model][parameter].value - value)/uncertainty for source, model, parameter, value, uncertainty in self.priors]

    def gaussLlh(self):
        resid = self.residualCalculation()
        return -0.5*resid*resid # Faster than **2

    def poissonLlh(self):
        model_calcs = self.f()
        returnvalue = self.temp_y * np.log(model_calcs) - model_calcs
        return returnvalue

    def llh(self, params, method='gaussian', emcee=False):
        methods = {'gaussian': self.gaussLlh, 'poisson': self.poissonLlh}
        self.setParameters(params)
        returnvalue = np.sum(methods[method.lower()]())
        if not np.isfinite(returnvalue):
            returnvalue = -1e99
        if not emcee:
            returnvalue *= -1
        return returnvalue

    def callback(self, params, iter, resid, *args, **kwargs):
        return None

    def residualCalculation(self):
        resid = self.resid()
        priors = self.gaussianPriors()
        if len(priors) > 0:
            resid = np.append(resid, priors)
        return resid

    def optimizeFunc(self, params):
        self.setParameters(params)
        return self.residualCalculation()

    def prepareFit(self):
        self.createParameters()
        self.createBounds()
        self.createLmParameters()

    def reportFit(self):
        return lm.fit_report(self.result)

    def fittingDifferenceCalculator(self, parameter_name, llh_selected=False, llh_method='gaussian', method='leastsq', kws={}, mcmc_kwargs={}, sampler_kwargs={}, filename=None):
        if parameter_name not in self.lmpars.keys():
            raise ValueError("Unknown parameter name {}".format(parameter_name))

        fit_kws = {'prepFit': False, 'llh_selected': llh_selected, 'llh_method': llh_method, 'method': method, 'kws': kws, 'mcmc_kwargs': mcmc_kwargs, 'sampler_kwargs': sampler_kwargs, 'filename': filename}
        needed_attr = 'chisqr'
        if llh_selected or llh_method == 'poisson':
            needed_attr = 'nllh_result'
        try:
            original_value = getattr(self, needed_attr)
        except AttributeError:
            self.fit(**fit_kws)
            original_value = getattr(self, needed_attr)

        copied_params = copy.deepcopy(self.pars)
        self.pars[parameter_name].vary = False

        # Define boundary calculating function
        def func(x):
            self.pars[parameter_name].value = x
            self.lmpars = self.pars
            self.fit(**fit_kws)
            value = getattr(self, needed_attr)
            return value - original_value
        return func

    def calculateUncertainties(self, parameter_name, llh_selected=False, llh_method='gaussian', method='leastsq', kws={}, mcmc_kwargs={}, sampler_kwargs={}, filename=None):
        if parameter_name not in self.lmpars.keys():
            raise ValueError("Unknown parameter name {}".format(parameter_name))

        fit_kws = {'prepFit': False, 'llh_selected': llh_selected, 'llh_method': llh_method, 'method': method, 'kws': kws, 'mcmc_kwargs': mcmc_kwargs, 'sampler_kwargs': sampler_kwargs, 'filename': filename}
        diff_calc = self.fittingDifferenceCalculator(parameter_name, **fit_kws)
        if llh_selected or llh_method == 'poisson':
            func_to_zero = lambda x: diff_calc(x) - 0.5
        else:
            func_to_zero = lambda x: diff_calc(x) - 1

    def fit(self, prepFit=True, llh_selected=False, llh_method='gaussian', method='leastsq', kws={}, mcmc_kwargs={}, sampler_kwargs={}, filename=None, steps=1000, nwalkers=50):
        self.temp_y = self.y()
        if prepFit:
            self.prepareFit()
        if llh_method.lower() == 'poisson':
            llh_selected = True

        kws = {}
        kwargs = {}
        if llh_selected or method.lower() == 'emcee':
            func = self.llh
            kws['method'] = llh_method
            if method.lower() in ['leastsq', 'least_squares']:
                method = 'nelder'
        else:
            func = self.optimizeFunc

        if method == 'emcee':
            func = self.llh
            kws['method'] = llh_method
            kws['emcee'] = True
            mcmc_kwargs['skip_initial_state_check'] = True
            if filename is not None:
                sampler_kwargs['backend'] = SATLASHDFBackend(filename)
            else:
                sampler_kwargs['backend'] = None

            kwargs = {'mcmc_kwargs': mcmc_kwargs,
                      'sampler_kwargs': sampler_kwargs}

            kwargs['sampler'] = SATLASSampler
            kwargs['steps'] = steps
            kwargs['nwalkers'] = nwalkers

        self.result = minimize(func, self.lmpars, method=method, iter_cb=self.callback, kws=kws, **kwargs)
        if llh_selected:
            self.llh_result = self.llh(self.result.params, method=llh_method)
            self.nllh_result = self.llh(self.result.params, method=llh_method)
        else:
            self.llh_result = None
            self.nllh_result = None
        del self.temp_y
        self.updateInfo()

    def readWalk(self, filename):
        reader = SATLASHDFBackend(filename)
        var_names = list(reader.labels)
        data = reader.get_chain(flat=False)
        try:
            self.result = SATLASMinimizer(self.llh, self.lmpars).process_walk(self.lmpars, data)
        except AttributeError:
            self.prepareFit()
            self.result = SATLASMinimizer(self.llh, self.lmpars).process_walk(self.lmpars, data)
        self.updateInfo()

    def updateInfo(self):
        self.lmpars = self.result.params
        self.setUncertainties(self.result.params)
        self.setParameters(self.result.params)
        self.nvarys = self.result.nvarys
        try:
            self.nfree = self.result.nfree
            self.ndata = self.result.ndata
            self.chisqr = self.result.chisqr
            self.redchi = self.result.redchi
        except:
            pass
        self.updateFitInfoSources()

    def updateFitInfoSources(self):
        for source_name, source in self.sources:
            source.nvarys = self.nvarys
            try:
                source.chisqr = self.chisqr
                source.ndata = self.ndata
                source.nfree = self.nfree
                source.redchi = self.redchi
            except:
                pass

    def toDataFrame(self):
        import pandas as pd
        row = []
        df = pd.DataFrame()
        for source_name, source in self.sources:
            data = {}
            row.append(source_name)
            p = self.pars[source_name]
            d = ()
            columns = []
            for model_name, model in source.models:
                pars = p[model_name]
                for parameter_name in pars.keys():
                    columns.extend([(model_name, parameter_name, 'Value'), (model_name, parameter_name, 'Uncertainty')])
                    d += (pars[parameter_name].value, pars[parameter_name].unc)
            columns.extend([(model_name, 'Fit quality', 'Chisquare'), (model_name, 'Fit quality', 'Reduced chisquare'), (model_name, 'Fit quality', 'NDoF')])
            d += (source.chisqr, source.redchi, source.nfree)
            data[source_name] = d
            d = pd.DataFrame.from_dict(data, orient='index')
            d.columns = pd.MultiIndex.from_tuples(columns)
            df = pd.concat([df, d])
        df.sort_index(axis=1, level=1, ascending=True, sort_remaining=False, inplace=True)
        return df

class Source:
    def __init__(self, x, y, xerr=None, yerr=1, name=None):
        super().__init__()
        self.x = x
        self.y = y
        self.xerr = xerr
        self.yerr_data = yerr
        if self.yerr_data == 1:
            self.yerr_data = np.ones(self.x.shape)
        if name is not None:
            self.name = name
        self.models = []

    def addModel(self, model, name=None):
        if name is None:
            name = model.name
        self.models.append((name, model))

    def params(self):
        params = {}
        for name, model in self.models:
            params[name] = model.params
        return params

    def f(self):
        for name, model in self.models:
            try:
                f += model.f(self.x)
            except UnboundLocalError:
                f = model.f(self.x)
        return f

    def evaluate(self, x):
        for name, model in self.models:
            try:
                f += model.f(x)
            except UnboundLocalError:
                f = model.f(x)
        return f

    def yerr(self):
        if not callable(self.yerr_data):
            return self.yerr_data
        else:
            return self.yerr_data(self.f())

class Model:
    def __init__(self, prefunc=None, name=None, pretransform=True):
        super().__init__()
        self.name = name
        self.prefunc = prefunc
        self.params = {}
        self.xtransformed = None
        self.xhashed = None

    def transform(self, x):
        if callable(self.prefunc):
            hashed = x.data.tobytes()
            if hashed == self.xhashed:
                x = self.xtransformed
            else:
                x = self.prefunc(x)
                self.xtransformed = x
                self.xhashed = hashed
        return x

    def setTransform(self, func):
        self.prefunc = func

    def params(self):
        return {}

    def setBounds(self, name, bounds):
        if name in self.params.keys():
            self.params[name].min = min(bounds)
            self.params[name].max = max(bounds)

    def setVary(self, name, vary):
        if name in self.params.keys():
            self.params[name].vary = vary

    def setExpr(self, name, expr):
        if name in self.params.keys():
            self.params[name].expr = expr

    def addParameter(self, **kwargs):
        name = kwargs.pop('name')
        self.params[name] = Parameter(**kwargs)

    def f(self, x):
        raise NotImplemented


class Parameter:
    def __init__(self, value=0, min=-np.inf, max=np.inf, vary=True, expr=None):
        super().__init__()
        self.value = value
        self.min = min
        self.max = max
        self.vary = vary
        self.expr = expr
        self.unc = None

    def __repr__(self):
        return '{}+/-{} ({} max, {} min, vary={})'.format(self.value, self.unc, self.max, self.min, self.vary)



from satlas2.core import Model, Parameter

import numpy as np

__all__ = ['ToFICR']

def ExtractionArray():
    a = 6.99797
    b = 6.9225642
    c = 0.0015824
    d = -0.0000079
    e = 3.2265697
    f = 0.0373329
    g = -0.0000951
    h = 10.0476155
    i = -0.0024371
    j = -0.0000381
    k = 26.6285599
    l = -0.090522
    m = 0.0000787
    n = 11.2570807
    o = -0.0326127
    p = 0.0000241
    q = 3.0220748
    r = -0.0069769
    s = 0.0000041
    
    array_b_e = np.array([
        [0,  2,  0.4,  a,  0,  0],
        [2,  4,  -0.68,  a,  0,  0],
        [4,  6,  -1.5,  a,  0,  0],
        [6,  8,  -2.1,  a,  0,  0],
        [8,  10,  -2.6,  a,  0,  0],
        [10,  15,  -3.2,  a,  0,  0],
        [15,  20,  -3.8,  a,  0,  0],
        [20,  30,  -4.2,  a,  0,  0],
        [30,  40,  -4.3,  a,  0,  0],
        [40,  60,  -4.4,  a,  0,  0],
        [60,  200,  -4.4,  b,  c,  d],
        [200,  230,  -4.4,  e,  f,  g],
        [230,  270,  -4.9,  e,  f,  g],
        [270,  300,  -5.4,  e,  f,  g],
        [300,  340,  -5.9,  h,  i,  j],
        [340,  370,  -6.5,  h,  i,  j],
        [370,  400,  -16.8,  h,  i,  j],
        [400,  440,  -180.9,  k,  l,  m],
        [440,  480,  -334.2,  k,  l,  m],
        [480,  510,  -271.5,  k,  l,  m],
        [510,  540,  -204.8,  k,  l,  m],
        [540,  600,  -836.7,  n,  o,  p],
        [600,  640,  -1500,  n,  o,  p],
        [640,  740,  -15300,  q,  r,  s],
        [740,  840,  -30000,  q,  r,  s],
        [840,  1600,  -30000,  0.06,  0,  0],
    ])
    return array_b_e

class ToFICR(Model):
    def __init__(self, f1=0, f2=0, f3=0, AxE=10, p=0.1, rm=1.413, rp=0.1, phase=0, mass=85, conversions=1.0, Ttot=400, Tf1=100, Tf2=300, ratio=25, ratio2=0, bkg=0.1, bkgratio=0, charge=1.0, profile="Konig", name=None):
        super().__init__(name=name)

        self.params = {'f1': Parameter(value=f1, vary=True),
                       'f2': Parameter(value=f2, vary=False),
                       'f3': Parameter(value=f3, vary=False),
                       'AxE': Parameter(value=AxE, vary=True),
                       'p': Parameter(value=p, vary=False),
                       'rm': Parameter(value=rm, vary=True),
                       'rp': Parameter(value=rp, vary=True),
                       'phase': Parameter(value=phase, vary=True),
                       'mass': Parameter(value=mass, vary=True),
                       'conversions': Parameter(value=conversions, vary=False),
                       'Ttot': Parameter(value=Ttot, vary=False),
                       'Tf1': Parameter(value=Tf1, vary=False),
                       'Tf2': Parameter(value=Tf2, vary=False),
                       'ratio': Parameter(value=ratio, vary=True, min=0, max=100),
                       'ratio2': Parameter(value=ratio2, vary=False, min=0, max=100),
                       'bkg': Parameter(value=bkg, vary=False),
                       'bkgratio': Parameter(value=bkgratio, vary=False, min=0, max=100),
                       'charge': Parameter(value=charge, vary=False),
                }
        self.mapping = {'Konig': self.konig,
                        'Double Konig': self.doublekonig,
                        'Triple Konig': self.triplekonig,
                        'Ideal Ramsey': self.ramsey,
                        'Damped Ramsey': self.dampedramsey,
                        'Asymmetrical Ramsey': self.asymmetricramsey}
        self.profiles = list(self.mapping.keys())
        self.profile = profile


    def _getTofFasterLoop(self, mu):
        q = self.params['charge'].value
        mass = self.params['mass'].value
        axial_energy = self.params['AxE'].value

        field_points = ExtractionArray()

        total_energy = axial_energy+q*field_points[0][2]+mu*field_points[0][3]

        EPSILON = 1.0e-30
        mu = np.array(mu)
        this_tof = np.zeros((len(field_points), mu.shape[0]))

        for k, i in enumerate(field_points):
            x1 = i[0]
            x2 = i[1]
            a = -mu*i[5]
            b = -mu*i[4]
            c = total_energy - q*i[2] - mu*i[3]
            determinant = b*b - 4 * a * c

            mask = np.logical_and.reduce([np.abs(a) > EPSILON, np.abs(b) > EPSILON])
            mask_posa_posdisc = np.logical_and.reduce([mask, a > 0, determinant > 0])
            mask_posa_negdisc = np.logical_and.reduce([mask, a > 0, determinant < 0])
            mask_nega_posdisc = np.logical_and.reduce([mask, a < 0, determinant > 0])
            mask_2 = np.logical_and.reduce([np.abs(a) <= EPSILON, np.abs(b) > EPSILON])
            mask_3 = np.logical_and.reduce([np.abs(a) <= EPSILON, np.abs(b) <= EPSILON])

            m = mask_posa_posdisc
            if m.sum() > 0:
                this_tof[k][m] += 1/np.sqrt(a[m]) * (    #Schaum 17.13.1
                              np.log(2 * np.sqrt(a[m]) * np.sqrt(a[m] * np.power(x2,2) + b[m] * x2 + c[m]) + 2 * a[m] * x2 + b[m])
                             -np.log(2 * np.sqrt(a[m]) * np.sqrt(a[m] * np.power(x1,2) + b[m] * x1 + c[m]) + 2 * a[m] * x1 + b[m])
                            )
            m = mask_posa_negdisc
            if m.sum() > 0:
                this_tof[k][m] = 1 / np.sqrt(a[m]) * (
                              np.arcsinh((2 * a[m] * x2 + b[m]) / np.sqrt(4 * a[m] * c[m] - b[m] * b[m]))
                             -np.arcsinh((2 * a[m] * x1 + b[m]) / np.sqrt(4 * a[m] * c[m] - b[m] * b[m]))
                              )
            m = mask_nega_posdisc
            if m.sum() > 0:
                this_tof[k][m] = - 1 / np.sqrt(-a) * (
                                  np.arcsin((2 * a[m] * x2 + b[m]) / np.sqrt(determinant[m]))
                                 -np.arcsin((2 * a[m] * x1 + b[m]) / np.sqrt(determinant[m]))
                                )
            m = mask_2
            if m.sum() > 0:
                this_tof[k][m] = 2 / np.sqrt(b[m]) * (
                              np.sqrt(b[m] * x2 + c[m])
                             -np.sqrt(b[m] * x1 + c[m])
                            )
            m = mask_3
            if m.sum() > 0:
                this_tof[k][m] = 1 / np.sqrt(c[m]) * (x2 - x1)
        time_of_flight = this_tof.sum(axis=0)
        return time_of_flight * np.sqrt(mass / 2) / 10

    def _getTofFaster(self, mu):
        q = self.params['charge'].value
        mass = self.params['mass'].value
        axial_energy = self.params['AxE'].value

        field_points = ExtractionArray()

        total_energy = axial_energy+q*field_points[0][2]+mu*field_points[0][3]

        EPSILON = 1.0e-30
        mu = np.array(mu)
        this_tof = np.zeros((len(field_points), mu.shape[0]))

        mu.shape = (1, mu.shape[0])
        x1 = np.ones(mu.shape) * np.atleast_2d(field_points[:, 0]).T
        x2 = np.ones(mu.shape) * np.atleast_2d(field_points[:, 1]).T
        a = -mu*np.atleast_2d(field_points[:, 5]).T
        b = -mu*np.atleast_2d(field_points[:, 4]).T

        c = total_energy - q*np.atleast_2d(field_points[:, 2]).T - mu*np.atleast_2d(field_points[:, 3]).T

        determinant = b*b - 4 * a * c

        mask = np.logical_and.reduce([np.abs(a) > EPSILON, np.abs(b) > EPSILON])
        mask_posa_posdisc = np.logical_and.reduce([mask, a > 0, determinant > 0])
        mask_posa_negdisc = np.logical_and.reduce([mask, a > 0, determinant < 0])
        mask_nega_posdisc = np.logical_and.reduce([mask, a < 0, determinant > 0])
        mask_2 = np.logical_and.reduce([np.abs(a) <= EPSILON, np.abs(b) > EPSILON])
        mask_3 = np.logical_and.reduce([np.abs(a) <= EPSILON, np.abs(b) <= EPSILON])

        m = mask_posa_posdisc
        if m.sum() > 0:
            this_tof[m] = 1/np.sqrt(a[m]) * (    #Schaum 17.13.1
                          np.log(2 * np.sqrt(a[m]) * np.sqrt(a[m] * np.power(x2[m],2) + b[m] * x2[m] + c[m]) + 2 * a[m] * x2[m] + b[m])
                         -np.log(2 * np.sqrt(a[m]) * np.sqrt(a[m] * np.power(x1[m],2) + b[m] * x1[m] + c[m]) + 2 * a[m] * x1[m] + b[m])
                        )
        m = mask_posa_negdisc
        if m.sum() > 0:
            this_tof[m] = 1 / np.sqrt(a[m]) * (
                          np.arcsinh((2 * a[m] * x2[m] + b[m]) / np.sqrt(4 * a[m] * c[m] - b[m] * b[m]))
                         -np.arcsinh((2 * a[m] * x1[m] + b[m]) / np.sqrt(4 * a[m] * c[m] - b[m] * b[m]))
                          )
        m = mask_nega_posdisc
        if m.sum() > 0:
            this_tof[m] = - 1 / np.sqrt(-a[m]) * (
                              np.arcsin((2 * a[m] * x2[m] + b[m]) / np.sqrt(determinant[m]))
                             -np.arcsin((2 * a[m] * x1[m] + b[m]) / np.sqrt(determinant[m]))
                            )
        m = mask_2
        if m.sum() > 0:
            this_tof[m] = 2 / np.sqrt(b[m]) * (
                          np.sqrt(b[m] * x2[m] + c[m])
                         -np.sqrt(b[m] * x1[m] + c[m])
                        )
        m = mask_3
        if m.sum() > 0:
            this_tof[m] = 1 / np.sqrt(c[m]) * (x2[m] - x1[m])
        time_of_flight = this_tof.sum(axis=0)
        return time_of_flight * np.sqrt(mass / 2) / 10

    def _magneticMoment(self, x, peak=1):
        conversions = self.params['conversions'].value
        time_tot = self.params['Ttot'].value*1e-3
        nu_c = self.params['f{:.0f}'.format(peak)].value
        mass = self.params['mass'].value
        pressure = self.params['p'].value*1e-7
        rad_minus0 = self.params['rm'].value*1e-3
        rad_plus0 = self.params['rp'].value*1e-3
        phase = self.params['phase'].value / 360 * 2 * np.pi

        # Phase is in radians
        omega_c = nu_c*2.*np.pi
        omega_rf = x*2.*np.pi

        omega_minus = 1700*2*np.pi
        omega_plus = (omega_c-omega_minus)

        k0 = conversions / time_tot *np.pi
        delta = pressure * omega_c / (1000 * 7 * 18.8e-4)
        gamma = delta / (omega_plus - omega_minus)

        omega_b1 = 0.5*np.sqrt(np.power((omega_rf-omega_c),2)+np.power((gamma*omega_c-k0),2))
        omega_b2 = 0.5*np.sqrt(np.power((omega_rf-omega_c),2)+np.power((gamma*omega_c+k0),2))
        omega_b = np.sqrt(omega_b1*omega_b2)

        theta1 = np.arctan((omega_rf-omega_c)/(gamma*omega_c-k0))
        theta1 = theta1*(theta1 >= 0) + (theta1+np.pi)*(theta1 < 0)

        theta2 = np.arctan((omega_rf-omega_c)/(gamma*omega_c+k0))
        theta2 = theta2*(theta2 >= 0) + (theta2+np.pi)*(theta2 < 0)

        theta = 0.5*(theta1+theta2)
        x = omega_b*time_tot
        y = 0.5*(omega_rf-omega_c)
        l = np.cosh(x*np.cos(theta))*np.cos(x*np.sin(theta)) #Re[np.cosh(xe^(i*theta))]
        m = np.sinh(x*np.cos(theta))*np.sin(x*np.sin(theta)) #Im[np.cosh(xe^(i*theta))]
        n = np.sinh(x*np.cos(theta))*np.cos(x*np.sin(theta)) #Re[np.sinh(xe^(i*theta))]
        p = np.cosh(x*np.cos(theta))*np.sin(x*np.sin(theta)) #Im[np.sinh(xe^(i*theta))]
        q1 = 0.5/omega_b*(rad_plus0 *(gamma*omega_c*np.cos(theta)+2.0*y*np.sin(theta))+rad_minus0*k0*(np.cos(phase)*np.cos(theta)+np.sin(phase)*np.sin(theta)))
        q2 = 0.5/omega_b*(rad_minus0*(gamma*omega_c*np.cos(theta)+2.0*y*np.sin(theta))+rad_plus0 *k0*(np.cos(phase)*np.cos(theta)-np.sin(phase)*np.sin(theta)))

        r1 = 0.5/omega_b*(rad_plus0 *(-gamma*omega_c*np.sin(theta)+2.0*y*np.cos(theta))+rad_minus0*k0*(-np.cos(phase)*np.sin(theta)+np.sin(phase)*np.cos(theta)))
        r2 = 0.5/omega_b*(rad_minus0*(-gamma*omega_c*np.sin(theta)+2.0*y*np.cos(theta))+ rad_plus0*k0*(-np.cos(phase)*np.sin(theta)-np.sin(phase)*np.cos(theta)))

        rad_plus_RE = np.exp(-delta*time_tot*0.5)*(np.cos(y*time_tot)*(rad_plus0 *l-(q1*n-r1*p))-np.sin(y*time_tot)*(rad_plus0 *m-(r1*n+p*q1)))
        rad_minus_RE = np.exp(-delta*time_tot*0.5)*(np.cos(y*time_tot)*(rad_minus0*l+(q2*n-r2*p))-np.sin(y*time_tot)*(rad_minus0*m+(r2*n+p*q2)))

        rad_plus_IM = np.exp(-delta*time_tot*0.5)*(np.sin(y*time_tot)*(rad_plus0 *l-(q1*n-r1*p))+np.cos(y*time_tot)*(rad_plus0 *m-(r1*n+p*q1)))
        rad_minus_IM = np.exp(-delta*time_tot*0.5)*(np.sin(y*time_tot)*(rad_minus0*l+(q2*n-r2*p))+np.cos(y*time_tot)*(rad_minus0*m+(r2*n+p*q2)))

        rad_minus = np.sqrt(np.power(rad_minus_RE,2.0) + np.power(rad_minus_IM,2.0))
        rad_plus = np.sqrt(np.power(rad_plus_RE, 2.0) + np.power(rad_plus_IM, 2.0))
        energy_minus = 0.5 * mass * 1.660520e-27 / 1.602177e-19 * np.power((rad_minus * omega_minus), 2.0)
        energy_plus = 0.5 * mass * 1.660520e-27 / 1.602177e-19 * np.power((rad_plus * omega_plus), 2.0)
        energy = energy_minus + energy_plus
        return energy #Returns radial energy in eV

    def _ramsey(self, x, peak=1):
        conversions = self.params['conversions'].value
        time_fringe = self.params['Tf1'].value * 1e-3
        time_tot = self.params['Ttot'].value * 1e-3
        nu_c = self.params['f{:.0f}'.format(peak)].value
        mass = self.params['mass'].value
        rad_minus0 = self.params['rm'].value * 1e-3

        time_wait = time_tot-2.*time_fringe
        omega_c =  nu_c*2.*np.pi
        omega_rf = x*2.*np.pi
        delta = omega_rf-omega_c
        g = conversions /time_fringe*np.pi*250./1000.
        omega_R = np.sqrt(np.power(2.*g,2.0)+np.power(delta,2.0))
        conversion = np.power(2.*g/omega_R *np.sin(omega_R*time_fringe/2.),2.0)*np.power(2.*np.cos(delta*time_wait/2.)*np.cos(omega_R*time_fringe/2.)-2.*delta/omega_R*np.sin(delta*time_wait/2.)*np.sin(omega_R*time_fringe/2.),2.0)
        omega_plus = omega_c
        energy_plus  = conversion*0.5*mass*1.660520e-27/1.602177e-19*np.power((rad_minus0*omega_plus),2.)
        return energy_plus

    def _ramseydamp(self, x, peak=1):
        conversions = self.params['conversions'].value
        time_fringe = self.params['Tf1'].value * 1e-3
        time_tot = self.params['Ttot'].value * 1e-3
        nu_c = self.params['f{:.0f}'.format(peak)].value
        mass = self.params['mass'].value
        pressure = self.params['p'].value * 1e-7
        rad_minus0 = self.params['rm'].value * 1e-3
        rad_plus0 = self.params['rp'].value * 1e-3
        phase = self.params['phase'].value / 360 * 2 * np.pi

        time_wait = (time_tot) - 2. * (time_fringe)
        omega_c =  (nu_c)*2.*np.pi
        omega_rf = (x)*2.*np.pi
        omega_minus = 2.*np.pi*180
        omega_plus = omega_c - omega_minus
        detuning = x - nu_c
        # Coupling constant
        g = conversions/(time_fringe) *np.pi*250. /1000.     #a_rf in units of "conversion"
        g_p2 = np.power(g, 2.0)
        gamma_p2 = np.power((pressure), 2.0)
        detuning_dash = detuning + 1j * 2 * pressure
        # Rabi frequency of the interconversion
        omega_R_dash = (np.sqrt(np.power(detuning_dash, 2.0) + 4.0 * g_p2))
        omega_R_dash_abs_p2 = np.power(abs(omega_R_dash), 2.0)
        d_p2 = 26.4e-3
        omega_z_p2 = (1.602177e-19 * 10) / (d_p2 * 1.660520e-27) / mass
        omega_1_p2 = np.power(omega_c, 2.0) - 2 * omega_z_p2
        
        M_SQRT1_2 = 1.0*10**(-8.5)
        gamma1_tilde = M_SQRT1_2* np.sqrt(np.sqrt(np.power(omega_1_p2 - 4 * gamma_p2, 2.0)+ 16 * gamma_p2 * np.power(omega_c, 2.0))- (omega_1_p2 - 4 * gamma_p2))
        conv_1 = np.cos((.5 * detuning * time_wait) + 1j * ( - pressure * time_wait))
        conv_2 = np.sin((.5 * detuning * time_wait) + 1j * ( - pressure * time_wait))
        ding = detuning_dash / omega_R_dash
        conversion = np.exp(-gamma1_tilde*(2.0*time_fringe+time_wait))*8.5*(4.0 * g_p2 / omega_R_dash_abs_p2)* np.power(abs(conv_1 * np.sin(omega_R_dash * time_fringe)+ ding * conv_2* (np.cos(omega_R_dash * time_fringe) - 1.0)),2.0)
        
        energy_plus = 0.5 * mass * 1.660520e-27 / 1.602177e-19 *conversion* np.power((rad_minus0 * omega_plus), 2.)
        return energy_plus

    def _ramseyasym(self, x, peak=1):
        conversions = self.params['conversions'].value
        time_fringe1 = self.params['Tf1'].value * 1e-3
        time_fringe2 = self.params['Tf2'].value * 1e-3
        time_tot = self.params['Ttot'].value * 1e-3
        nu_c = self.params['f{:.0f}'.format(peak)].value
        mass = self.params['mass'].value
        rad_minus0 = self.params['rm'].value * 1e-3
        rad_plus0 = self.params['rp'].value * 1e-3
        phase = self.params['phase'].value / 360 * 2 * np.pi

        time_wait = time_tot - time_fringe1 - time_fringe2
        omega_c =  nu_c*2.*np.pi
        omega_rf = x*2.*np.pi
        delta = omega_rf - omega_c
        g = conversions /((time_fringe1) + (time_fringe2))*(np.pi)*500./1000.
        omega_R = np.sqrt(np.power(2.*g ,2.0) + np.power(delta, 2.0))
        aa1=np.cos(delta*time_fringe1/2.0)*(np.sin(omega_R*(time_fringe1+time_fringe2)/2.0))
        aa2=delta/omega_R*(np.sin(delta*time_wait/2.0))*(np.cos(omega_R*(time_fringe1+time_fringe2)/2.0)-np.cos(omega_R*(time_fringe1 - time_fringe2)/2.))
        aa3=np.power(2.*g/omega_R*(np.sin(delta*time_wait/2.0))*(np.sin(omega_R*(time_fringe1+time_fringe2)/2.0)), 2.0)
        conversion = np.power(2.*g/omega_R, 2.0) *np.power(aa1 + aa2, 2.0) + aa3
        omega_plus = omega_c #Assumes w_+ ~ w_-
        energy_plus=conversion*0.5*mass*1.660520e-27/1.602177e-19*np.power((rad_minus0*omega_plus),2.)
        return energy_plus

    def konig(self, x):
        mu = self._magneticMoment(x, peak=1)
        tof = self._getTofFaster(mu/7)
        return tof

    def konigbkg(self, x):
        mu1 = self._magneticMoment(x, peak=1)
        tof1 = self._getTofFaster(mu1/7)
        tof2 = self.params['Tofcrap']
        ratio = self.params['ratio'].value / 100
        return tof2*ratio + tof1 * (1 - ratio)

    def doublekonig(self, x):
        mu1 = self._magneticMoment(x, peak=1)
        tof1 = self._getTofFaster(mu1/7)
        mu2 = self._magneticMoment(x, peak=2)
        tof2 = self._getTofFaster(mu2/7)
        ratio = self.params['ratio'].value / 100
        return tof2*ratio + tof1 * (1 - ratio)

    def triplekonig(self, x):
        mu1 = self._magneticMoment(x, peak=1)
        tof1 = self._getTofFaster(mu1/7)
        mu2 = self._magneticMoment(x, peak=2)
        tof2 = self._getTofFaster(mu2/7)
        mu3 = self._magneticMoment(x, peak=3)
        tof3 = self._getTofFaster(mu3/7)
        ratio1 = self.params['ratio'].value / 100
        ratio2 = self.params['ratio2'].value / 100
        return tof3 * ratio2 + tof2*ratio1 + tof1 * (1 - ratio1 - ratio2)

    def ramsey(self, x):
        mu = self._ramsey(x, peak=1)
        tof = self._getTofFaster(mu/7)
        return tof

    def dampedramsey(self, x):
        mu = self._ramseydamp(x, peak=1)
        tof = self._getTofFaster(mu/7)
        return tof

    def asymmetricramsey(self, x):
        mu = self._ramseyasym(x, peak=1)
        tof = self._getTofFaster(mu/7)
        return tof

    def f(self, x):
        x = self.transform(x)
        bkg = self.params['bkg'].value
        bkgratio = self.params['bkgratio'].value / 100
        return self.mapping[self.profile](x) * (1 - bkgratio) + bkg * bkgratio



from satlas2.core import Model, Parameter

import numpy as np
from scipy.special import wofz
from sympy.physics.wigner import wigner_6j, wigner_3j

__all__ = ['ExponentialDecay', 'Polynomial']

sqrt2 = 2 ** 0.5
sqrt2log2t2 = 2 * np.sqrt(2 * np.log(2))
log2 = np.log(2)

class Polynomial(Model):
    def __init__(self, p, name=None, prefunc=None):
        super().__init__(name=name, prefunc=prefunc)
        self.params = {'p'+str(len(p)-(i+1)): Parameter(value=P, min=-np.inf, max=np.inf, vary=True) for i, P in enumerate(p)}

    def f(self, x):
        x = self.transform(x)
        p = [self.params[paramkey].value for paramkey in self.params.keys()]
        return np.polyval(p, x)

class ExponentialDecay(Model):
    def __init__(self, a, tau, bkg, name=None, prefunc=None):
        super().__init__(name=name, prefunc=prefunc)
        self.params = {
            'amplitude': Parameter(value=a, min=-np.inf, max=np.inf, vary=True),
            'halflife': Parameter(value=tau, min=-np.inf, max=np.inf, vary=True),
            'background': Parameter(value=bkg, min=-np.inf, max=np.inf, vary=True),
        }

    def f(self, x):
        x = self.transform(x)
        a = self.params['amplitude'].value
        b = self.params['halflife'].value
        c = self.params['background'].value
        return a*np.exp(-log2*x/b)+c



from .core import *
from . import core

from satlas2.models import *
import satlas2.models

from .overwrite import *
from . import overwrite

from .plotting import *
from . import plotting

from . import _version
__version__ = _version.version



from satlas2.core import Model, Parameter

import numpy as np
from scipy.special import wofz
from sympy.physics.wigner import wigner_6j, wigner_3j

__all__ = ['Linear', 'Exponential', 'HFS']

sqrt2 = 2 ** 0.5
sqrt2log2t2 = 2 * np.sqrt(2 * np.log(2))
log2 = np.log(2)

class Linear(Model):
    def __init__(self, a, b, name=None, prefunc=None):
        super().__init__(name=name, prefunc=prefunc)
        self.params = {
                'a': Parameter(value=a, min=-np.inf, max=np.inf, vary=True),
                'b': Parameter(value=b, min=-np.inf, max=np.inf, vary=True),
                }

    def f(self, x):
        x = self.transform(x)
        a = self.params['a'].value
        b = self.params['b'].value
        return a*x+b

class Exponential(Model):
    def __init__(self, a, b, c, name=None, prefunc=None):
        super().__init__(name=name, prefunc=prefunc)
        self.params = {
            'amplitude': Parameter(value=a, min=-np.inf, max=np.inf, vary=True),
            'halflife': Parameter(value=b, min=-np.inf, max=np.inf, vary=True),
            'background': Parameter(value=c, min=-np.inf, max=np.inf, vary=True),
        }

    def f(self, x):
        x = self.transform(x)
        a = self.params['amplitude'].value
        b = self.params['halflife'].value
        c = self.params['background'].value
        return a*np.exp(log2*x/b)+c

class HFS(Model):
    def __init__(self, I, J, A=[0, 0], B=[0, 0], C=[0, 0], df=0, fwhm=50, bkg=1, name=None, N=None, offset=0, poisson=0, scale=1.0, racah=True, prefunc=None):
        super().__init__(name=name, prefunc=prefunc)
        J1, J2 = J
        lower_F = np.arange(abs(I - J1), I+J1+1, 1)
        upper_F = np.arange(abs(I - J2), I+J2+1, 1)

        self.lines = []
        self.intensities = {}
        self.scaling_Al = {}
        self.scaling_Bl = {}
        self.scaling_Cl = {}
        self.scaling_Au = {}
        self.scaling_Bu = {}
        self.scaling_Cu = {}

        for i, F1 in enumerate(lower_F):
            for j, F2 in enumerate(upper_F):
                if abs(F2 - F1) <= 1 and not F2 == F1 == 0.0:
                    if F1 % 1 == 0:
                        F1_str = '{:.0f}'.format(F1)
                    else:
                        F1_str = '{:.0f}_2'.format(2*F1)

                    if F2 % 1 == 0:
                        F2_str = '{:.0f}'.format(F2)
                    else:
                        F2_str = '{:.0f}_2'.format(2*F2)

                    line = '{}to{}'.format(F1_str, F2_str)
                    self.lines.append(line)

                    C1, D1, E1 = self.calcShift(I, J1, F1)
                    C2, D2, E2 = self.calcShift(I, J2, F2)

                    self.scaling_Al[line] = C1
                    self.scaling_Bl[line] = D1
                    self.scaling_Cl[line] = E1
                    self.scaling_Au[line] = C2
                    self.scaling_Bu[line] = D2
                    self.scaling_Cu[line] = E2

                    intens = float((2 * F1 + 1) * (2 * F2 + 1) * \
                         wigner_6j(J2, F2, I, F1, J1, 1.0) ** 2) # DO NOT REMOVE CAST TO FLOAT!!!
                    self.intensities['Amp'+line] = Parameter(value=intens, min=0, vary=not racah)

        norm = max([p.value for p in self.intensities.values()])
        for n, v in self.intensities.items():
            v.value /= norm

        pars = {'centroid': Parameter(value=df),
                'Al': Parameter(value=A[0]),
                'Au': Parameter(value=A[1]),
                'Bl': Parameter(value=B[0]),
                'Bu': Parameter(value=B[1]),
                'Cl': Parameter(value=C[0], vary=False),
                'Cu': Parameter(value=C[1], vary=False),
                'bkg': Parameter(value=bkg),
                'FWHMG': Parameter(value=fwhm, min=0.01),
                'FWHML': Parameter(value=fwhm, min=0.01),
                'scale': Parameter(value=scale, min=0, vary=racah)}
        if N is not None:
            pars['N'] = Parameter(value=N, vary=False)
            pars['Offset'] = Parameter(value=offset)
            pars['Poisson'] = Parameter(value=poisson, min=0, max=1)
            self.f = self.fShifted
        else:
            self.f = self.fUnshifted
        pars = {**pars, **self.intensities}

        self.params = pars

        if I < 2 or J1 < 2:
            self.params['Cl'].vary = False
        if I < 2 or J2 < 2:
            self.params['Cu'].vary = False
        if I < 1 or J1 < 1:
            self.params['Bl'].vary = False
        if I < 1 or J2 < 1:
            self.params['Bu'].vary = False
        if I == 0 or J1 == 0: 
            self.params['Al'].vary = False
        if I == 0 or J2 == 0:
            self.params['Au'].vary = False
        self.xtransformed = None
        self.xhashed = None

    def fUnshifted(self, x):
        centroid = self.params['centroid'].value
        Al = self.params['Al'].value
        Au = self.params['Au'].value
        Bl = self.params['Bl'].value
        Bu = self.params['Bu'].value
        Cl = self.params['Cl'].value
        Cu = self.params['Cu'].value
        FWHMG = self.params['FWHMG'].value
        FWHML = self.params['FWHML'].value
        scale = self.params['scale'].value
        bkg = self.params['bkg'].value

        result = np.zeros(len(x))
        x = self.transform(x)
        for line in self.lines:
            pos = centroid + Au * self.scaling_Au[line] + Bu * self.scaling_Bu[line] + Cu * self.scaling_Cu[line] - Al * self.scaling_Al[line] - Bl * self.scaling_Bl[line] - Cl * self.scaling_Cl[line]
            result += self.params['Amp' + line].value * self.peak(x - pos, FWHMG, FWHML)

        return scale * result + bkg

    def fShifted(self, x):
        centroid = self.params['centroid'].value
        Al = self.params['Al'].value
        Au = self.params['Au'].value
        Bl = self.params['Bl'].value
        Bu = self.params['Bu'].value
        FWHMG = self.params['FWHMG'].value
        FWHML = self.params['FWHML'].value
        scale = self.params['scale'].value
        N = self.params['N'].value
        offset = self.params['Offset'].value
        poisson = self.params['Poisson'].value
        bkg = self.params['bkg'].value

        result = np.zeros(len(x)) 
        for line in self.lines:
            pos = centroid + Au * self.scaling_Au[line] + Bu * self.scaling_Bu[line] + Cu * self.scaling_Cu[line] - Al * self.scaling_Al[line] - Bl * self.scaling_Bl[line] - Cl * self.scaling_Cl[line]
            for i in range(N + 1):
                if self.prefunc:
                    result += self.params['Amp' + line].value * self.peak(self.prefunc(x - i * offset) - pos, FWHMG, FWHML) * (poisson**i)/np.math.factorial(i)
                else:
                    result += self.params['Amp' + line].value * self.peak(x - pos - i * offset, FWHMG, FWHML) * (poisson**i)/np.math.factorial(i)

        return scale * result + bkg

    def peak(self, x, FWHMG, FWHML):
        z = self.preparePeak(x, FWHMG, FWHML)
        n = self.norm(FWHML, FWHMG)
        ret = wofz(z).real
        return ret/n

    def norm(self, FWHML, FWHMG):
        return wofz(1j * FWHML / (FWHMG * sqrt2)).real

    def preparePeak(self, x, FWHMG, FWHML):
        sigma, gamma = FWHMG / sqrt2log2t2, FWHML / 2
        z = (x + 1j * gamma) / (sigma * sqrt2)
        return z

    def calcShift(self, I, J, F):
        phase = (-1)**(I+J+F)
        contrib = []
        for k in range(1, 4):
            n = float(wigner_6j(I, J, F, J, I, k))
            d = float(wigner_3j(I, k, I, -I, 0, I) * wigner_3j(J, k, J, -J, 0, J))
            shift = phase * n / d
            if not np.isfinite(shift):
                contrib.append(0)
            else:
                if k == 1:
                    shift = shift * (I*J)
                elif k == 2:
                    shift = shift / 4
                contrib.append(shift)
        return contrib

    def pos(self):
        centroid = self.params['centroid'].value
        Al = self.params['Al'].value
        Au = self.params['Au'].value
        Bl = self.params['Bl'].value
        Bu = self.params['Bu'].value
        Cl = self.params['Cl'].value
        Cu = self.params['Cu'].value
        pos = []
        for line in self.lines:
            pos.append(centroid + Au * self.scaling_Au[line] + Bu * self.scaling_Bu[line] + Cu * self.scaling_Cu[line] - Al * self.scaling_Al[line] - Bl * self.scaling_Bl[line] - Cl * self.scaling_Cl[line])
        return pos
